unit StringConv;

interface

uses SysUtils;

const
	CRLF : String = #13#10;
	ssSkip = 1;
	ssScan = 0;

type
	PInteger = ^Integer;
	EStrConvError = class(Exception)
  public
  	ErrorCode : Integer;
	end;


function IntMax( L1,L2 : Integer ) : Integer; assembler;
function RoundCents( const ACurr : Currency ) : Currency;
function IncMonth( ADate : TDateTime; const AMonthes : Integer ) : TDateTime;

function SpellCurrency( ANum : Currency ) : String;
function SpellDate( const ADate : TDateTime ) : String;

function GetTokenInteger(var ptr:PChar):Integer;
function GetIntegers(ptr:PChar;Arg:Array of PInteger):PChar;

function Comma( const str1, str2 : string ) : string;
function CommaAppend( var str1 : string; const str2 : string ) : string;
function LineAppend( var str1 : string; const str2 : string ) : string;
function StrOpStr( const str1, op, str2 : string ) : string;
function StrSpaceStr( const str1, op, str2 : string ) : string;
function StrAppendOpStr( var str1 : string; const op, str2 : string ) : string;

function SkipBlanks(ptr:PChar):PChar;
function SkipBlanksLen( var ptr : PChar; var len : Integer ):PChar;
function FindLexLen( var ptr : PChar; var len : Integer ):PChar;
function NextLine( var ptr : PChar; var len : Integer ) : PChar;
function NextTokenLen( var ptr : PChar; var len : Integer ) : PChar;
function NextToken( var ptr : PChar ) : PChar;

function Quote( const Line : String ) : String;
function XQuote( const Line, Table : String ) : String;
function ConditionalQuote( const Line, Table : String ) : String;
function NullQuote( const str : String ) : String;
function TextToCSV( Src : String ) : String;
function UnQuoteToken(var ptr:PChar):String;
function QStrSplit( var str : String; c : char ) : String;

function GetTerm( var ptr:PChar; var len:Integer ) : String;
function GetTermEx( var ptr:PChar; var len:Integer; ADelimitors:String; const AGetDelimitor:Boolean ) : String;
function GetInteger(var ptr:PChar; var len:Integer) : Integer;
function GetIntegerDef( var ptr : PChar; var len : Integer; const ADefaultValue : Integer ) : Integer;
function GetString( var ptr:PChar; var len:Integer ) : String;
function GetLine( var ptr : PChar; var len : Integer ) : String;
function GetKey( var ptr:PChar; var len:Integer ) : String;
function GetSubStr( var ptr : PChar; var len : Integer; const ATerminators : String ) : String;
function GetNumVariant( var ptr:PChar; var len:Integer ) : Variant;
function QStrScan( str : String; c : char ) : Integer;
function QMemScan( ptr : PChar; len : Integer; c : Char ) : PChar;
function QSkipChars( ptr:PChar; len:Integer; chars : String )	: Integer;
function QScanChars( ptr:PChar; len:Integer; chars : String ) : Integer;
function QStrUnquote( str : String ) : String;
{function UnQuote(var ptr:PChar):String;}
function DeEsc( Str : String ) : String;
function StripCRLF( Str : String ) : String;
function MacroExpand( src:String; Args : Array of String; c : Char ):String;
//function EncodeId( Origin, Id :Integer ) : String;
//procedure DecodeId( var Origin,Id : Integer; ptr : PChar ); assembler;
function c2x( Dest,Src : Pointer; len : Integer ) : Integer;
function i2x( Dest, Src : Pointer; len : Integer ) : Integer; assembler;
function BinToHex( const ptr : Pointer; len : Integer ) : String;
function DumpToBuf( Dest, Src : Pointer; len : Integer ) : Integer; assembler;
function dump( ptr : Pointer; len : Integer ) : String;
procedure AsmXLAT( Str : String; const sFrom, sTo : String; Pad : Char ); assembler;
function Translate( const Str, sFrom, sTo : String; Pad : Char ) : String;

function MemScan(Str: PChar; Len : LongInt; Chr : Char): PChar; assembler;
function MemCpy( Dest : Pointer; Src : Pointer; Len : Integer ) : Pointer; assembler;
function MemCmp( Src1 : Pointer; Src2 : Pointer; Len : Integer ) : Integer; assembler;

function WStrComp( src1, src2 : PWideChar ):Integer;

function lstrcmpi( const Str1, Str2 : PChar ) : LongInt;
	stdcall;  external 'kernel32.dll' name 'lstrcmpi';
function lstrcmp( const Str1, Str2 : PChar ) : LongInt;
	stdcall;  external 'kernel32.dll' name 'lstrcmp';

const
	AlphaNum : String = '$@abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890';
	Printable: String = ' !%&()-+=*[]{}~;:"?/\<>.,_|#$@abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890';
	Delimitors : String='=,.<>;:\/?+-*&^%(){}[]|!~ '#9#10#13#0;
	i2xtable : array [0..15] of char = '0123456789abcdef';

//=============================================================
implementation
uses	Windows;

//=============================================================
function IntMax( L1,L2 : Integer ) :Integer; assembler;
asm
	cmp	eax,edx
	jge	@@Ret
	mov	eax,edx
@@Ret:
end;

//-------------------------------------------------------------
               { EAX         EDX            ECX }
function MemScan(Str: PChar; Len : LongInt; Chr : Char): PChar; assembler;
asm
	push    edi
	push    eax      { Str }
	mov     al,chr   { CL }
	mov     ecx,len  { EDX }
	pop     edi      { Str }
	repne   scasb
	mov     eax,0
	jne     @@1
	mov     eax,edi
	dec     eax
@@1:
	pop     edi
end;

               { EAX            EDX            ECX }
function MemCpy( Dest: Pointer; Src : Pointer; Len : Integer ): Pointer; assembler;
asm
	push    edi
	push		esi
	mov			esi,edx
	mov			edi,eax
	rep			movsb
	mov     eax,edi
	pop     esi
	pop			edi
end;

               { EAX            EDX             ECX }
function MemCmp( Src1: Pointer; Src2 : Pointer; Len : Integer ): Integer; assembler;
asm
  push    edi
  push		esi
  mov			esi,edx
  mov			edi,eax
  xor			eax,eax
  repe		cmpsb
  je			@@2
  dec			edi
  dec			esi
  mov     al,[edi]
  sub			al,[esi]
@@2:
	pop     esi
	pop			edi
end;

{---------------------------------------------------------------------------}
function WStrComp( src1, src2 : PWideChar ):Integer; assembler;
asm
	or		eax,eax
  jnz		@@1
	or		edx,edx
  jz		@@Ret
  dec		eax
  jmp		@@Ret

@@1:
	or		edx,edx
  jnz		@@2
  xor		eax,eax
  inc		eax
  jmp		@@Ret

@@2:
  push	edi
  push	esi
	mov		esi,eax
  mov		edi,edx
  xor		eax,eax
  xor		ecx,ecx
  dec		ecx
  repne	scasw
  not		ecx
  mov		edi,edx
	xor		edx,edx
  repe	cmpsw
	mov		ax,[esi-2]
	mov		dx,[edi-2]
	sub		eax,edx
	pop		esi
  pop		edi
@@Ret:
end;

//-------------------------------------------------------------
function TextToCSV( Src : String ) : String;
const	MAXLEN = 256;
var buf : Array [0..(MAXLEN*2+1)] of Char;
		ptr : PChar;
    Len : Integer;
	function TextToBufCSV( src, buf : PChar; len : Integer ) : Integer; assembler;
	asm
		test	eax,eax	// Src
  	jz		@@Ret
  	test	ecx,ecx	// len
  	jz		@@RetNIL
		push	edi
		push	esi
		push	ebx

		push	edx
  	mov		esi,eax	// Src
		mov		ebx,ecx
  	mov		al,'"'
		mov		[edx],al
		inc   edx
		mov		edi,edx
	@@1:
		mov		edi,esi
	repnz	scasb
		jnz		@@7
		mov		edi,edx	// Buf
		xchg	ecx,ebx
		sub		ecx,ebx
		jz		@@5
	rep	movsb
	@@5:
		stosb		//  
		mov		edx,edi
		mov		ecx,ebx
		test	ebx,ebx
		jnz		@@1
		jmp		@@9

@@RetNIL:
		xor		eax,eax
		jmp		@@Ret

@@7:
		mov		edi,edx	// Buf
		mov		ecx,ebx
	rep	movsb
@@9:
		stosb
		pop		edx
		mov		eax,edi
		pop		ebx
		pop		esi
		pop		edi
		sub		eax,edx
@@Ret:
	end;

begin
  len := Length(Src);
	if (len=0)or((MemScan(PChar(Src),len,'"')=nil)and(MemScan(PChar(Src),len,';')=nil)) then Result := Src
	else if len>MAXLEN then
		begin
			GetMem( ptr, (len shl 1)+2 );
			len := TextToBufCSV( PChar(Src), ptr, len );
			Result := Copy(ptr,1,len);
			FreeMem( ptr );
		end
	else
		begin
			len := TextToBufCSV( PChar(Src), buf, len );
			Result := Copy(buf,1,len);
		end;
end; {TextToCSV}
//-------------------------------------------------------------
// !!! buf   ,    len*3+2 !!!
function asmXquote( const Src : String; buf : PChar; const Tbl : String ) : Integer;
var	Table : Array [0..255] of Char;
asm
	push	edi
	push	esi
	push	ebx
	push	eax

  test	ecx,ecx
  jz		@@DefaultTable
  mov   esi,ecx
	mov		ecx,[esi-4]
	test	ecx,ecx
	jnz		@@BuildTable
@@DefaultTable:
	lea		esi,Printable
	mov		ecx,[esi-4]
@@BuildTable:
//  
  lea		ebx,Table
  mov		edi,ebx
  mov		ecx,256
  mov		eax,1
  rep		stosb
	mov		ecx,[esi-4]
@@NextChar:
  lodsb
  mov		[ebx+eax].Byte,0
	loop	@@NextChar
  mov		[ebx+''''].Byte,2
  mov		[ebx+'$'].Byte,2
//   
  pop		esi
	push	edx
  mov		edi,edx
	xor		eax,eax
	mov		ecx,[esi-4]
	xor		edx,edx
	mov		byte ptr [edi],''''
	inc		edi
	jcxz	@@End
// 
@@Next:
	lodsb
	mov		dl,al
  xlat
  jmp		[@@Vector+eax*4]

@@Vector:
	dd		@@Skip
  dd		@@Escape
  dd		@@Double

@@Escape:
	mov		byte ptr [edi],'$'
	inc		edi
	mov		al,dl
	shr		al,4
	and		dl,15
	mov		al,byte ptr[i2xtable+eax]
	mov   [edi],al
	inc		edi
	mov		al,byte ptr[i2xtable+edx]
	mov   [edi],al
	inc		edi
	loop	@@Next
	jmp		@@End

@@Double:
	mov		[edi],dl
	inc		edi
@@Skip:
	mov		[edi],dl
	inc		edi
	loop	@@Next
@@End:
	mov		byte ptr [edi],''''
	inc		edi
	pop		edx
	mov		eax,edi
	sub		eax,edx
	pop		ebx
	pop		esi
	pop		edi
end; {asmXquote}
//-------------------------------------------------------------
function XQuote( const Line, Table : String ) : String;
const	MAXLEN = 256;
var buf : Array [0..(MAXLEN*3+1)] of Char;
		ptr : PChar;
    len : Integer;
begin
	len := Length(Line);
	if len=0 then Result := ''''''
	else if len>MAXLEN then
		begin
			GetMem( ptr, len*3+2 );
			len := asmXquote( Line, ptr, Table  );
			Result := Copy(ptr,1,len);
			FreeMem( ptr );
		end
	else
		begin
			len := asmXquote( Line, buf, Table );
			Result := Copy(buf,1,len);
		end;
end; {XQuote}
//-------------------------------------------------------------
function asmquote( src, buf : PChar; len : Integer ): Integer; assembler;
asm
{	test	eax,eax
  jz		@@Ret
  test	ecx,ecx   ecx<>0 <-   
  jnz		@@Start
	xor		eax,eax
	jmp		@@Ret
@@Start:}
	push	edi
	push	esi
	push	ebx
	push	edx
	mov		esi,eax
	mov		edi,edx
	mov		ebx,ecx
	mov		al,''''
	stosb
@@Loop:
	mov		edx,edi
	mov		edi,esi
	repne	scasb
	mov   ecx,edi
	sub		ecx,esi
	sub		ebx,ecx
	mov		edi,edx
	rep		movsb
	stosb
	mov		ecx,ebx
	test	ebx,ebx
	jnz		@@Loop
	pop		edx
	mov		eax,edi
	sub		eax,edx
	pop		ebx
	pop   esi
	pop		edi
@@Ret:
end;
//-------------------------------------------------------------
function Quote( const Line : String ) : String;
const	MAXLEN = 256;
var buf : Array [0..(MAXLEN*2+1)] of Char;
		ptr : PChar;
    len : Integer;
begin
	len := Length(Line);
	if len=0 then Result := ''''''
	else if len>MAXLEN then
		begin
			GetMem( ptr, len*2+2 );
			len := asmquote( PChar(Line), ptr, len );
			Result := Copy(ptr,1,len);
			FreeMem( ptr );
		end
	else
		begin
			len := asmquote( PChar(Line), buf, len );
			Result := Copy(buf,1,len);
		end;
end; {Quote}
//-------------------------------------------------------------
function NullQuote( const str : String ) : String;
begin
	if str='' then Result := 'NULL' else	Result := Quote(str);
end;

{---------------------------------------------------------------------------}
//                       EAX  EDX          ECX
function asmUnQuote( var src; buf : PChar; len : Integer ): Integer; assembler;
asm
	OR		EAX,EAX
  JZ		@@Ret
	PUSH	ESI
	PUSH	EDI
  PUSH	src
  MOV		EDI,buf
  MOV		ESI,[src]
  OR		ESI,ESI
  JZ		@@RetNIL
	LODSB
  CMP		AL,' '
  JC		@@RetNIL
  CMP		AL,''''
  JE		@@1	{    }
  CMP		AL,'"'
  JNE		@@UNQ	{         }
{    }
@@1:  MOV		AH,AL	{    }
@@2:	LODSB
	CMP		AL,' '
  JC		@@Unmatch
  CMP		AL,AH
  JE		@@End
	CMP		AL,'$'
  JE		@@ESC
@@Store:
	STOSB
	LOOP	@@2
@@Overflow:	{  -   }
	MOV		EAX,-1
  JMP		@@Abort

@@Unmatch:	{  -      }
	MOV		EAX,-3
  JMP		@@Abort

@@ESC:
	LODSB
  CMP		AL,'$'
  JE		@@Store
  SUB		AL,'0'
  JC		@@ErrEsc
  CMP		AL,32
  JC		@@Store
@@ErrEsc:
	MOV		EAX,-2	{  -    # }
@@Abort:
	XOR		ESI,ESI	{ nil }
	JMP		@@Exit

@@NextUnq:
	LODSB
@@UNQ:
	CMP		AL,' '
  JC		@@DEC
	CMP		AL,','
  JE    @@DEC
	STOSB
  LOOP  @@NextUnq
  JMP		@@OverFlow

@@RetNil:
	XOR		EAX,EAX
	POP		EDI		{src}
	JMP		@@Ret0

@@End:
  LODSB
	CMP		AL,AH
	JE		@@Store
@@DEC:
	DEC		ESI
@@Result:
	MOV		EAX,EDI
  SUB		EAX,EDX
@@Exit:
	POP		EDI		{src}
	MOV		[EDI],ESI
@@Ret0:
  POP		EDI
	POP		ESI
@@Ret:
end;

{---------------------------------------------------------------------------}
                      { EAX           EDX }
function QStrUnquoteBuf( src : String; buf : PChar ): Integer; assembler;
asm
	OR		EAX,EAX
  JZ		@@Ret
	PUSH	ESI
	PUSH	EDI
  MOV		EDI,EAX
  OR		EDI,EDI
  JZ		@@RetNIL
  MOV		ECX,[EDI-4]
  OR		ECX,ECX
  JZ		@@RetNIL
//    
	MOV		AX,2920h
@@SkipBlanks:
  REPE	SCASB
	JE		@@RetNIL
  DEC		EDI
  INC		ECX
	XOR		AL,AH
  CMP		AL,[EDI]
  JE		@@SkipBlanks
//   //
	MOV		ESI,EDI
  MOV		EDI,EDX	{buf}
//  
@@NextUnq:
	LODSB
@@Unquoted:
  CMP		AL,''''
  JE		@@Quoted
  CMP		AL,'"'
  JE		@@Quoted
	STOSB
  LOOP  @@NextUnq
	JMP		@@Exit

@@Quoted:
  MOV		AH,AL	{   -    }
	LOOP	@@Dequote
  JMP		@@UnMatch

//   
@@Dequote:
	LODSB
  CMP		AL,AH
  JE		@@Quote
	CMP		AL,'#'
  JNE		@@Store
//    
  DEC		ECX
  JCXZ	@@ErrESC
	LODSB
  CMP		AL,'#'
  JE		@@Store
  SUB		AL,'0'
  JC		@@ErrESC
  CMP		AL,32
  JC		@@Store
@@ErrESC:
	MOV		EAX,-2	{  -    # }
	JMP		@@Ret0

//       
@@Quote:
  DEC		ECX
  JCXZ	@@Exit
  LODSB
  CMP		AL,AH		//    /
  JNE		@@Unquoted
@@Store:
	STOSB
@@Iterate:
	LOOP	@@Dequote
@@Unmatch:	{  -      }
	MOV		EAX,-3
  JMP		@@Ret0

@@RetNil:
	XOR		EAX,EAX
	JMP		@@Ret0

@@Exit:
	MOV		EAX,EDI
  SUB		EAX,EDX
@@Ret0:
  POP		EDI
	POP		ESI
@@Ret:
end;

function QStrUnquote( str : String ) : String;
const	MAXLEN = 1024;
var buf : Array [0..(MAXLEN-1)] of Char;
		ptr : PChar;
    Len : Integer;
    Msg : String;
begin
  Len := Length(str);
  Result := '';
  if Len>MAXLEN then
	 begin
  	GetMem( ptr, Len );
		Len := QStrUnquoteBuf( str, ptr );
    if Len>0 then Result := Copy( ptr, 1, Len);
    FreeMem( ptr );
   end
  else
   begin
		Len := QStrUnquoteBuf( str, buf );
    Result := Copy( buf, 1, Len);
	 end;
	if Len<0 then
  begin
  	case -Len of
    2: Msg := 'StrUnquote:    #';
    3: Msg := 'StrUnquote:    ';
    else
    	Msg := 'StrUnquote:   '+IntToStr(Len);
    end;
    raise EStrConvError.Create( Msg );
  end;
end;


{---------------------------------------------------------------------------}
function UnQuote( var ptr : PChar ) : String;
const	MAXLEN=1024;
var	Len : Integer;
		buf : Array [0..(MAXLEN-1)] of Char;
    Msg : String;
begin
	ptr := SkipBlanks(ptr);
  Len := asmUnQuote(ptr,buf,MAXLEN);
  if Len<=0 then
	 begin
  	Result := '';
    if Len<0 then
    begin
	    ptr := nil;
	  	case -Len of
  	  1: Msg := 'UnQuote:  ';
    	2: Msg := 'UnQuote:    #';
	    3: Msg := 'UnQuote:    ';
  	  else
    		Msg := 'Unquote:   '+IntToStr(Len);
	    end;
  	  raise EStrConvError.Create( Msg );
    end;
	 end
  else	Result := Copy(buf,1,Len);
end;

{---------------------------------------------------------------------------}
function UnQuoteToken( var ptr : PChar ) : String;
begin
	Result := UnQuote( ptr );
  NextToken( ptr );
end;

{---------------------------------------------------------------------------}
{                  EAX  EDX          ECX }
function asmDeEsc( src, buf : PChar; len : Integer ): Integer; assembler;
asm
	OR		EAX,EAX
  JZ		@@Ret
  OR		ECX,ECX
  JZ		@@Ret0
	PUSH	ESI
	PUSH	EDI
  MOV		EDI,buf
  MOV		ESI,src
	MOV		AH,'#'
@@2:	LODSB
	CMP		AL,AH
  JNE		@@Store
	LODSB
  LOOP	@@3
@@ErrEsc:
	XOR		EAX,EAX
  DEC		EAX  {  -    # }
	JMP		@@Exit

@@Ret0:
	XOR		EAX,EAX
  JMP		@@Ret

@@3:
	CMP		AL,AH
  JE		@@Store
  SUB		AL,'0'
  JC		@@ErrEsc
  CMP		AL,32
  JNC		@@ErrEsc
@@Store:
	STOSB
	LOOP	@@2
	MOV		EAX,EDI
  SUB		EAX,EDX
@@Exit:
  POP		EDI
	POP		ESI
@@Ret:
end;

{---------------------------------------------------------------------------}
function DeEsc( Str : String ) : String;
const	MAXLEN=1024;
var	Len : Integer;
		buf : Array [0..(MAXLEN-1)] of Char;
    ptr : PChar;

function DeEscBuf( src, ptrbuf : PChar; Len : Integer ) : String;
var	Msg : String;
begin
  Len := asmDeEsc(PChar(Str),ptrbuf,Length(Str));
  if Len<=0 then
	 begin
  	Result := '';
    if Len<0 then
    begin
	  	if Len =-1 then Msg := 'DeEscBuf:    #'
  	  else Msg := 'DeEscBuf:   '+IntToStr(Len);
  	  raise EStrConvError.Create( Msg );
		end;
	 end
	else if Len = Length(Str) then Result:=Str
  else	Result := Copy(ptrbuf,1,Len);
end;

begin
	if Str='' then Result:=''
  else
	 begin
	  Len := Length(Str);
	  if Len>MAXLEN then
  	 begin
  		GetMem( ptr, Len );
			Result := DeEscBuf(PChar(Str),ptr,Len);
  	  FreeMem( ptr );
		 end
		else Result := DeEscBuf(PChar(Str),buf,Len);
	 end;
end;


{---------------------------------------------------------------------------}
{                      EAX  EDX          ECX }
function asmStripCRLF( src, buf : PChar; len : Integer ): Integer; assembler;
asm
	OR		EAX,EAX
  JZ		@@Ret
  OR		ECX,ECX
  JZ		@@Ret0
	PUSH	ESI
	PUSH	EDI
  MOV		EDI,buf
  MOV		ESI,src
	XOR		AH,AH
{   CRLF }
@@FindCRLF:
	LODSB
	CMP		AL,13
  JE		@@CR
	CMP		AL,10
  JE		@@CR
@@Store:
  STOSB
  LOOP	@@FindCRLF
	JMP		@@Exit

@@StoreCR:
  MOV		BYTE PTR[EDI],13
  INC		EDI
  MOV		BYTE PTR[EDI],10
  INC		EDI
  JMP   @@Store

@@Ret0:
	XOR		EAX,EAX
  JMP		@@Ret

{     }
@@SkipCRLF:
	LODSB
  CMP		AL,10
  JE		@@CR
	CMP		AL,13
  JNE		@@StoreCR
@@CR:
  LOOP	@@SkipCRLF

@@Exit:
	MOV		EAX,EDI
  SUB		EAX,EDX
  POP		EDI
	POP		ESI
@@Ret:
end;

{---------------------------------------------------------------------------}
function StripCRLF( Str : String ) : String;
const	MAXLEN=1024;
var	Len : Integer;
		buf : Array [0..(MAXLEN-1)] of Char;
    ptr : PChar;

function StripCRLFBuf( src, ptrbuf : PChar; Len : Integer ) : String;
begin
  Len := asmStripCRLF(PChar(Str),ptrbuf,Length(Str));
  if Len=0 then Result := ''
	else if Len = Length(Str) then Result:=Str
  else	Result := Copy(ptrbuf,1,Len);
end;

begin
	if Str='' then Result:=''
  else
	 begin
	  Len := Length(Str);
	  if Len>MAXLEN then
  	 begin
  		GetMem( ptr, Len );
			Result := StripCRLFBuf(PChar(Str),ptr,Len);
  	  FreeMem( ptr );
		 end
		else Result := StripCRLFBuf(PChar(Str),buf,Len);
	 end;
end;

{---------------------------------------------------------------------------}
{function EncodeId( Origin, Id :Integer ) : String;
const table : array [0..63] of char =
	'@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_ abcdefghijklmnopqrstuvwxyz;<=>?';
var Res : Array [0..7] of Char;

function Encode6( val : Integer; ptr : PChar ):Pchar; assembler;
asm
  PUSH	EDI
	PUSH	EBX
  LEA		EBX,table
	MOV		ECX,4
  MOV		EDI,EDX
	MOV		EDX,EAX
  SHL		EDX,8
@@Next:
  SHLD	EAX,EDX,6
  SHL		EDX,6
  AND		EAX,63
	XLAT
  STOSB
  LOOP	@@Next
  MOV		EAX,EDI
	POP		EBX
	POP		EDI
end;

begin
  Encode6( Id, Encode6( Origin, Res ));
	Result := Copy( Res, 1, 8 );
end;}
{---------------------------------------------------------------------------}
//   hex  
function c2x( Dest, Src : Pointer; len : Integer ) : Integer; assembler;
asm
	push	esi
	push	edi
  push	eax
	mov		esi,edx
	mov		edi,eax
	xor		eax,eax
	xor		edx,edx
	mov		al,'0'
	stosb
	mov		al,'x'
	stosb
@@loop:
	lodsb
	mov		dl,al
	shr		al,4
	and		dl,15
	mov		al,byte ptr[i2xtable+eax]
	mov   [edi],al
	inc		edi
	mov		al,byte ptr[i2xtable+edx]
	mov   [edi],al
	inc		edi
	loop	@@loop
	mov		eax,edi
  pop		edx
  sub		eax,edx
	pop		edi
	pop		esi
end;
{---------------------------------------------------------------------------}
//   hex   
function i2x( Dest, Src : Pointer; len : Integer ) : Integer; assembler;
const	i2xtable : array [0..15] of char = '0123456789abcdef';
asm
	push	esi
	push	edi
  push	eax
	mov		esi,edx
  add		esi,ecx
  dec		esi
	mov		edi,eax
	xor		eax,eax
	xor		edx,edx
	mov		al,'0'
	stosb
	mov		al,'x'
	stosb
	std
@@loop:
	lodsb
	mov		dl,al
	shr		al,4
	and		dl,15
	mov		al,byte ptr[i2xtable+eax]
	mov   [edi],al
	inc		edi
	mov		al,byte ptr[i2xtable+edx]
	mov   [edi],al
	inc		edi
	loop	@@loop
	cld
	mov		eax,edi
  pop		edx
  sub		eax,edx
	pop		edi
	pop		esi
end;
//-------------------------------------------------------------
function BinToHex( const ptr : Pointer; len : Integer ) : String;
begin
	SetLength(Result,(len+1)shl 1);
	c2x(PChar(Result),ptr,len);
end;
//-------------------------------------------------------------
//  
//   = (len*3)+((len-1) shr 2)-1
function DumpToBuf( Dest, Src : Pointer; len : Integer ) : Integer; assembler;
const	i2xtable : array [0..15] of char = '0123456789abcdef';
asm
	push	esi
	push	edi
  push	eax
	push	ebx
	mov		esi,edx
	mov		edi,eax
	xor		eax,eax
	xor		ebx,ebx
	xor		edx,edx
	jmp		@@enter
@@loop:
	inc		bl
	mov		al,bl
	and		al,15
	jnz   @@1
	mov		word ptr [edi],$0a0d
	inc		edi
	jmp		@@2
@@1:
	mov		byte ptr [edi],' '
	inc		edi
	and		al,3
	jnz   @@enter
	mov		byte ptr [edi],' '
@@2:
	inc		edi
@@enter:
	lodsb
	mov		dl,al
	shr		al,4
	and		dl,15
	mov		al,byte ptr[i2xtable+eax]
	mov   [edi],al
	inc		edi
	mov		al,byte ptr[i2xtable+edx]
	mov   [edi],al
	inc		edi
	loop	@@loop
	mov		eax,edi
	pop		ebx
  pop		edx
  sub		eax,edx
	pop		edi
	pop		esi
end;
//-------------------------------------------------------------
function dump( ptr : Pointer; len : Integer ) : String;
begin
	if (len=0)or(ptr=nil) then Result := ''
	else
		begin
			SetLength(Result,(len*3)+((len-1) shr 2)-1);
			DumpToBuf(Pointer(Result),ptr,len);
		end;
end;
//-------------------------------------------------------------
{function EncodeId( Origin, Id :Integer ) : String;
var	i64 : int64;
		buf : array [0..18] of char;
begin
	i64.Lo := Id; i64.Hi := Origin;
	SetString( Result,	buf, i2x(@buf,@i64,8) );
end;}
//-------------------------------------------------------------
{procedure DecodeId( var Origin,Id : Integer; ptr : PChar ); assembler;
asm
  PUSH	ESI
  MOV		ESI,ECX
	PUSH	EDX
  PUSH	EAX
	MOV		ECX,4
@@Next:
	LODSB
  SHL		EDX,6
  AND		EAX,63
  OR		EDX,EAX
  LOOP	@@Next
  POP		EAX
  MOV		DWORD PTR[EAX],EDX
	MOV		ECX,4
@@Next2:
	LODSB
  SHL		EDX,6
  AND		EAX,63
  OR		EDX,EAX
  LOOP	@@Next2
  POP		EAX
  MOV		DWORD PTR[EAX],EDX
	POP		ESI
end;}

{---------------------------------------------------------------------------}
function asmMacroExpand(
	src : String; dest : PChar; destlen : Integer;
	args : Array of string;  c : Char ) : Integer;
asm
	PUSH	ESI
  PUSH	EDI
  MOV		ESI,EAX		{src}
  OR		ESI,ESI
  JZ		@@Exit
	MOV		EDI,EDX 	{dest}
  OR		EDI,EDI
  JZ		@@BadParm
  PUSH	EDX
  MOV		EDX,ECX		{destlen}
  MOV		ECX,[ESI-4]	{length(src)}
  OR		ECX,ECX
  JZ		@@Ret0
  SUB		EDX,ECX
  JC		@@Overflow
  MOV		EAX,args
  OR		EAX,EAX
  JNZ		@@Start
@@BadParm:
  XOR		EAX,EAX
	DEC		EAX
  JMP		@@Exit

@@SkipLoop:
	LODSB
  CMP		AL,''''
  JE    @@Store
@@StoreQuote:
  STOSB
	LOOP	@@SkipLoop
  JMP		@@EndLoop

@@Start:
  MOV		AH,c
@@Loop:
	LODSB
  CMP		AL,''''
  JE		@@StoreQuote
  CMP		AL,AH
  JNE		@@Store
	INC		EDX
  DEC		ECX
  JECXZ @@EndLoop
  LODSB
  CMP		AL,AH
  JE		@@Store
  CMP		AL,'9'+1
  JNC		@@Store
  CMP		AL,'0'
  JC		@@Store
  SUB		AL,'0'
	PUSH	ESI
  PUSH	ECX
	MOVZX	ECX,AL
  CMP		ECX,args.Integer[-4]
  JA		@@BadNum
	MOV		ESI,args
	MOV		ESI,[ESI+ECX*4]
  OR		ESI,ESI
  JZ		@@Continue
  MOV		ECX,[ESI-4]
	OR		ECX,ECX
	JZ		@@Continue
	SUB		EDX,ECX
  JC		@@Overflow
  REP		MOVSB
@@Continue:
  POP		ECX
  POP		ESI
  MOV		AH,c
  JMP		@@Iterate

@@BadNum:
	POP		ECX
  POP		ESI
	XOR		EAX,EAX
	DEC		EAX
	DEC		EAX
	DEC		EAX
	JMP		@@Exit

@@Ret0:
	POP		EAX
  XOR		EAX,EAX
  JMP		@@Exit

@@Overflow:
  POP		EAX
  XOR		EAX,EAX
	DEC		EAX
	DEC		EAX
  JMP		@@Exit

@@Store:
  STOSB
@@Iterate:
	LOOP	@@Loop
@@EndLoop:
	MOV		EAX,EDI
  POP		EDX
  SUB		EAX,EDX
@@Exit:
	POP		EDI
  POP		ESI
end;

function MacroExpand( src: String; Args : Array of String; c : Char ):String;
const	MAXLEN=4096;
var	buf : Array [0..(MAXLEN-1)] of Char;
		len : Integer;
    Msg : String;
begin
  if src='' then Result := ''
  else
    begin
    	len := asmMacroExpand( src, buf, MAXLEN, Args, c );
    	if len<=0 then
        begin
    	 	  case -Len of
          1: Msg := 'MacroExpand:  ';
          2: Msg := 'MacroExpand:  ';
         	3: Msg := 'MacroExpand:   ';
    		  else
		    	  Msg := 'MacroExpand:   '+IntToStr(Len);
      		end;
		      raise EStrConvError.Create( Msg );
    	  end
      else Result := Copy( buf, 1, len );
    end;
end;

{---------------------------------------------------------------------------}
function StrOpStr( const str1, op, str2 : string ) : string;
begin
	if (str1<>'')and(str2<>'') then Result := '('+str1+')'+op+'('+str2+')'
  else
		if str1<>'' then Result := str1 else Result := str2;
end;
{---------------------------------------------------------------------------}
function StrAppendOpStr( var str1: string; const op, str2 : string ) : string;
begin
	if (str1<>'')and(str2<>'') then
	 begin
		if PChar(str1)^='(' then	str1 := str1+op+'('+str2+')'
		else str1 := '('+str1+')'+op+'('+str2+')';
	 end
  else
		if str2<>'' then str1 := str2;
	Result := str1;
end;
{---------------------------------------------------------------------------}
function StrSpaceStr( const str1, op, str2 : string ) : string;
begin
	if (str1<>'')and(str2<>'') then Result := str1+op+str2
  else
		if str1<>'' then Result := str1 else Result := str2;
end;
{---------------------------------------------------------------------------}
function Comma( const str1, str2 : string ) : string;
begin
	if (str1<>'')and(str2<>'') then Result := str1+','+str2
  else
		if str1<>'' then Result := str1 else Result := str2;
end;
{---------------------------------------------------------------------------}
function CommaAppend( var str1: string; const str2 : string ) : string;
begin
	if (str1<>'')and(str2<>'') then	str1 := str1+','+str2
  else if str2<>'' then str1 := str2;
	Result := str1;
end;
{---------------------------------------------------------------------------}
function LineAppend( var str1: string; const str2 : string ) : string;
begin
	if (str1<>'')and(str2<>'') then	str1 := str1+#13#10+str2
  else if str2<>'' then str1 := str2;
	Result := str1;
end;
{---------------------------------------------------------------------------}
function NextTokenLen( var ptr : PChar; var len : Integer ) : PChar; assembler;
asm
	OR		EAX,EAX
  JZ		@@Ret
  PUSH	EDI
  PUSH	EAX
  MOV		EDI,[EAX]
  OR		EDI,EDI
  JZ    @@RetNil
  MOV		ECX,[EDX]
  OR		ECX,ECX
  JZ		@@RetNil
	MOV		AX,', '
@@Skip:
  REPE	SCASB
	JE		@@RetNIL
	JCXZ	@@RetNIL
  JNC		@@Skip
  CMP		AH,[EDI-1]
  JE		@@Skip2
	JMP		@@RetEDI

@@RetNil:
	XOR		EAX,EAX
  MOV		[EDX],EAX
  POP		ECX
  MOV		[ECX],EAX
 	JMP		@@Ret0

@@Skip2:
  REPE	SCASB
	JE		@@RetNIL
	JCXZ	@@RetNIL
  JNC		@@Skip2
@@RetEDI:
	DEC		EDI
  INC		ECX
	MOV		EAX,EDI
  POP		EDI
  MOV		[EDI],EAX
  MOV		[EDX],ECX
@@Ret0:
	POP		EDI
@@Ret:
end;
{---------------------------------------------------------------------------}
function SkipBlanks( ptr : PChar ):PChar;assembler;
asm
	OR		EAX,EAX
  JZ		@@Ret
	PUSH	ESI
  MOV		ESI,EAX
	XOR		EAX,EAX
@@SkipSpaces:
	LODSB
	CMP		AL,' '
  JE		@@SkipSpaces
  CMP		AL,9
  JE		@@SkipSpaces
  OR		EAX,EAX
  JZ		@@RetNil
  DEC		ESI
	MOV		EAX,ESI
@@RetNil:
	POP		ESI
@@Ret:
end;
{---------------------------------------------------------------------------}
function SkipBlanksLen( var ptr : PChar; var len : Integer ):PChar;assembler;
asm
	test	EAX,EAX
  JZ		@@Ret
	PUSH	EDI
  PUSH	EAX
	test	ECX,ECX
  MOV		EDI,[EAX]
  test	EDI,EDI
  JZ		@@RetNil
  test	EDX,EDX
  JZ		@@RetNil
  MOV		ECX,[EDX]
  test	ECX,ECX
  JLE		@@RetNil
	MOV		AX,2920h
@@SkipBlanks:
  REPE	SCASB
	JE		@@RetNIL
  DEC		EDI
  INC		ECX
	XOR		AL,AH
  CMP		AL,[EDI]
  JE		@@SkipBlanks
	JMP		@@RetEDI

@@RetNIL:
	XOR		EDI,EDI
@@RetEDI:
	POP		EAX
  XCHG	EAX,EDI
  MOV		[EDI],EAX
  MOV		[EDX],ECX
	POP		EDI
@@Ret:
end;

{---------------------------------------------------------------------------}
function FindLexLen( var ptr : PChar; var len : Integer ) : PChar; assembler;
asm
	test	EAX,EAX
  JZ		@@Ret
	PUSH	EDI
  PUSH	EAX
	XOR		ECX,ECX
  MOV		EDI,[EAX]
  test	EDI,EDI
  JZ		@@RetNil
  test	EDX,EDX
  JZ		@@RetNil
  MOV		ECX,[EDX]
  test	ECX,ECX
  JLE		@@RetNil
	MOV		EAX,20h
@@Skip:
  REPE	SCASB
	JE		@@RetNIL
	JCXZ	@@RetNIL
  JNC		@@Skip
  DEC		EDI
  INC		ECX
	JMP		@@RetEDI

@@RetNIL:
	XOR		EDI,EDI
@@RetEDI:
	POP		EAX
  XCHG	EAX,EDI
  MOV		[EDI],EAX
  MOV		[EDX],ECX
	POP		EDI
@@Ret:
end;

{---------------------------------------------------------------------------}
function Translate( const Str, sFrom, sTo : String; Pad : Char ) : String;
begin
	if (Str='')or(sFrom='') then Result := Str
	else
	 begin
		SetString( Result, PChar(Str), Length(Str) );
    AsmXLAT( Result, sFrom, sTo, Pad );
	 end;
end;

// !!!    !!!
procedure AsmXLAT( Str : String; const sFrom, sTo : String; Pad : Char );
var	Table : Array [0..255] of Char;
asm
	OR		EAX,EAX
  JZ		@@Ret
  OR		EDX,EDX
  JZ		@@Ret
	PUSH	ESI
	PUSH	EDI
  PUSH	EBX
	MOV		ESI,[EAX-4]
  OR		ESI,ESI
  JZ		@@Ret0
	MOV		ESI,[EDX-4]
  OR		ESI,ESI
  JZ		@@Ret0
//  
	PUSH	EAX
  PUSH	ECX
  LEA		EBX,table
  MOV		EDI,EBX
  XOR		EAX,EAX
@@FillTBL1:
  STOSB
  INC		AL
  JNZ 	@@FillTBL1
  POP		EDI
  MOV		ESI,EDX
  MOV		EDX,[EDX-4]
  OR		EDI,EDI
  JZ		@@FillTBL3
  MOV		ECX,[EDI-4]
  SUB		EDX,ECX
  JC		@@100
  MOV		ECX,[ESI-4]
  XOR		EDX,EDX
@@100:
	OR		ECX,ECX
  JZ		@@FillTBL3
@@FillTBL2:
  LODSB
	LEA		EBX,Table.[EAX]
  MOV		AL,[EDI]
	MOV		[EBX],AL
  INC		EDI
	LOOP	@@FillTBL2
@@FillTBL3:
  LEA		EBX,Table
	MOV		ECX,EDX
  OR		ECX,ECX
  JZ		@@FillTBL9
	MOV		DL,Pad
  XOR		EAX,EAX
@@FillTBL4:
  LODSB
	MOV		[EAX+EBX],DL
	LOOP	@@FillTBL4
//  
@@FillTBL9:
	POP		EDI
  MOV		ECX,[EDI-4]
@@NextChar:
	MOV		AL,[EDI]
	XLAT
	STOSB
  LOOP	@@NextChar
@@Ret0:
	POP		EBX
	POP		EDI
  POP		ESI
@@Ret:
end;

{---------------------------------------------------------------------------}
{  op=0 - Scan, op<>0 - Skip }
                   { EAX        EDX          ECX }
function QScanSkipChars( ptr:PChar; len:Integer; chars : String; op:Integer ) : Integer;
var	Table : Array [0..255] of Char;
asm
	OR		EAX,EAX
  JZ		@@Ret
  OR		EDX,EDX
  JZ		@@RetNIL
	PUSH	ESI
	PUSH	EDI
  PUSH	EBX
  OR		ECX,ECX
  JZ		@@Ret0
  MOV   ESI,ECX
	MOV		ECX,[ESI-4]
  OR		ECX,ECX
  JZ		@@Ret0
//  
	PUSH	EAX
  LEA		EBX,table
  MOV		EDI,EBX
  MOV		ECX,256
  XOR		EAX,EAX
  REP		STOSB
  MOV		[EBX+''''].byte,2
  MOV		[EBX+'"'].byte,2
	MOV		ECX,[ESI-4]
@@NextChar:
  LODSB
  MOV		[EBX+EAX].byte,1
	LOOP	@@NextChar
  POP		ESI		{ptr}
  MOV		EDI,ESI
  MOV		ECX,EDX
  LEA		EDX,@@VectorScan
  MOV		EAX,op
  OR		EAX,EAX
  JZ    @@Next
  LEA		EDX,@@VectorSkip
  CMP		[EBX+''''].byte,1
  JNE		@@1
  MOV		[EBX+''''].byte,2
	JMP		@@2
@@1:  MOV		[EBX+''''].byte,0
@@2:  CMP		[EBX+'"'].byte,1
  JNE		@@3
  MOV		[EBX+'"'].byte,2
  JMP		@@Next
@@3:  MOV		[EBX+'"'].byte,0
@@Next:
	LODSB
  XLAT
  JMP		[EDX].Pointer[EAX*4]

@@VectorSkip:
	DD		@@Exit
  DD		@@IterateSkip
  DD		@@SkipQuoted

@@VectorScan:
  DD		@@IterateScan
	DD		@@Exit
  DD		@@SkipQuoted

@@RetNIL:
  XOR		EAX,EAX
  JMP		@@Ret

@@IterateScan:
  LOOP	@@Next
@@Ret0:
  XOR		EAX,EAX
	JMP		@@RetEAX

@@SkipQuoted:
	DEC		ECX
	JCXZ	@@Ret0
  DEC		ESI
  LODSB
  XCHG	EDI,ESI
	REPNE	SCASB
  XCHG	ESI,EDI
  JNE		@@Ret0
  JMP		@@Next

@@IterateSkip:
  LOOP	@@Next
	INC		ESI
@@Exit:
	MOV		EAX,ESI
	SUB		EAX,EDI
@@RetEAX:
	POP		EBX
	POP		EDI
  POP		ESI
@@Ret:
end;

function QScanChars( ptr:PChar; len:Integer; chars : String )	: Integer;
begin
	Result := QScanSkipChars(ptr,len,chars,0);
end;
function QSkipChars( ptr:PChar; len:Integer; chars : String )	: Integer;
begin
	Result := QScanSkipChars(ptr,len,chars,1);
end;

//-------------------------------------------------------------
function ConditionalQuote( const Line, Table : String ) : String;
var len : Integer;
begin
  len := Length(Line);
	if len=0 then Result := ''''''
	else if QScanSkipChars(PChar(Line),len,AlphaNum,1)>len then Result := Line
	else Result := XQuote(Line,Table);
end;

//-------------------------------------------------------------
//  ,       const Delimitors,      
// Quoted-    
function GetTerm( var ptr:PChar; var len:Integer ) : String;
var	i : Integer;
begin
	Result := '';
	if SkipBlanksLen(ptr,len)=nil then Exit;
	if (ptr^='''')or(ptr^='"') then
  begin
		i := QScanSkipChars( ptr, len, '''"', 1 ); {skip}
  	if i>0 then
			begin
				Dec(i);
				if i=0 then Exit;
				Result := QStrUnquote(Copy(ptr,1,i));
				Inc(ptr,i);
				Dec(len,i)
			end
  	else
	  	begin
  	  	Result := QStrUnquote(Copy(ptr,1,len));
    		ptr := nil;
	    	len := 0;
			end;
		Exit;
	end;
	i := QScanSkipChars( ptr, len, Delimitors, 0 );{scan}
  if i>0 then
		begin
			Dec(i);
			if i<>0 then
      begin
  			Result := Copy(ptr,1,i);
	  		Inc(ptr,i);
		  	Dec(len,i)
      end;
		end
  else
		begin
			Result := Copy(ptr,1,len);
			ptr := nil;
			len := 0;
		end;
end;
//-------------------------------------------------------------
//  ,       ADelimitors,      
//   quoted-! -     
function GetTermEx( var ptr:PChar; var len:Integer; ADelimitors : String;
	const AGetDelimitor : Boolean ) : String;
var	i : Integer;
begin
	Result := '';
	if ADelimitors='' then ADelimitors := Delimitors;
	if SkipBlanksLen(ptr,len)=nil then Exit;
	i := QScanSkipChars( ptr, len, ADelimitors, 0 );{scan}
  if i>0 then
   begin
    Dec(i);
    if i=0 then
    begin
    	if AGetDelimitor then
	    begin
  	  	Result := ptr^;
    	  Inc(ptr);
      	Dec(len);
			end;
      Exit;
    end;
		Result := Copy(ptr,1,i);
    Inc(ptr,i);
    Dec(len,i)
   end
  else
   begin
    Result := Copy(ptr,1,len);
    ptr := nil;
    len := 0;
	 end;
end;
{---------------------------------------------------------------------------}
function GetKey( var ptr:PChar; var len:Integer ) : String;
const	Delimitors:string='=\/'#9#10#13#0;
var	i,k : Integer;
		p : PChar;
begin
	Result := '';
	if SkipBlanksLen(ptr,len)=nil then Exit;
	i := QScanSkipChars( ptr, len, Delimitors, 0 );{scan}
  if i>0 then
   begin
    Dec(i);
    if i=0 then Exit;
   end
  else i := len;
	Dec(len,i);
	k := i;
	p := ptr+i-1;
	while (k>0)and(p^=' ') do
	begin
		Dec(p);
		Dec(k);
	end;
	if k>0 then Result := Copy(ptr,1,k);
  Inc(ptr,i);
end;
{---------------------------------------------------------------------------}
function GetString( var ptr:PChar; var len:Integer ) : String;
var	i : Integer;
begin
	Result := '';
	if SkipBlanksLen(ptr,len)=nil then Exit;
  if (ptr^='''')or(ptr^='"') then
  begin
		i := QScanSkipChars( ptr, len, '''"', 1 ); {skip}
  	if i>0 then
	   begin
  	  Dec(i);
    	if i=0 then Exit
			else Result := QStrUnquote(Copy(ptr,1,i));
  	  Inc(ptr,i);
    	Dec(len,i)
	   end
  	else
	   begin
  	  Result := QStrUnquote(Copy(ptr,1,len));
    	ptr := nil;
	    len := 0;
		 end;
		Exit;
	end;
	i := QScanSkipChars( ptr, len, ',; '#9#10#13#0, 0 ); {scan}
  if i>0 then
	 begin
		Dec(i);
    if i=0 then Exit
		else Result := Copy(ptr,1,i);
	  Inc(ptr,i);
   	Dec(len,i)
   end
 	else
   begin
 	  Result := Copy(ptr,1,len);
   	ptr := nil;
    len := 0;
	 end;
end;
//-------------------------------------------------------------
//  , Result =      
function GetSubStr( var ptr : PChar; var len : Integer; const ATerminators : String ) : String;
type	PWord = ^Word;
var	i : Integer;
begin
	if (ptr=nil)and(len<=0) then
	begin
		Result := '';
		Exit;
	end;
	i := QScanSkipChars( ptr, len, ATerminators, 0 ); {scan}
  if i<1 then
	begin
		Result := Copy(ptr,1,len);
		ptr := nil;
		len := 0;
		Exit;
	end;
	Dec(len,i);
	Dec(i);
	if i>0 then Result := Copy(ptr,1,i) else Result :='';
	Inc(ptr,i+1);
end;
//-------------------------------------------------------------
const ScanCRLF : String = #10#13#0;
//-------------------------------------------------------------
//   -   CRLF.
// ptr,len    CRLF
function GetLine( var ptr : PChar; var len : Integer ) : String;
type	PWord = ^Word;
var	i : Integer;
begin
	if (ptr=nil)and(len<=0) then
	begin
		Result := '';
		Exit;
	end;
	i := QScanSkipChars( ptr, len, ScanCRLF, 0 ); {scan}
  if i<1 then
	begin
		Result := Copy(ptr,1,len);
		ptr := nil;
		len := 0;
		Exit;
	end;
	Dec(len,i);
	Dec(i);
	if i>0 then Result := Copy(ptr,1,i) else Result :='';
	Inc(ptr,i);
	if PWord(ptr)^=$0A0D then
	begin
		Inc(ptr);
		Dec(len);
	end;
	Inc(ptr);
end;
//-------------------------------------------------------------
function NextLine( var ptr : PChar; var len : Integer ) : PChar;
label ExitLen0, ExitNil;
var	i : Integer;
begin
	if ptr=nil then goto ExitLen0;
	i := QScanSkipChars( ptr, len, ScanCRLF, 0 ); {scan}
  if i<=0 then goto ExitLen0;
	Inc(ptr,i);
 	Dec(len,i);
	if len<=0 then goto ExitLen0;
	if (ptr^=#10)or(ptr^=#13) then
	begin
		Inc(ptr);
 		Dec(len);
	end;
	if len<=0 then goto ExitLen0;
	Result := ptr;
	Exit;

ExitLen0:
	len := 0;
ExitNil:
	ptr := nil;
	Result := nil;
end;

//-------------------------------------------------------------
function GetInteger(var ptr:PChar;var len:Integer):Integer;
var n:Integer;
begin
  if (SkipBlanksLen(ptr,len)=nil)or(ptr^<' ') then Result := 0
  else
   begin
    Val(ptr,Result,n);
		if n>0 then
		begin
			Dec(n);
    	Inc(ptr,n);
			Dec(len,n);
		end;
   end;
end;
//-------------------------------------------------------------
function GetIntegerDef( var ptr : PChar; var len : Integer; const ADefaultValue : Integer ) : Integer;
var n:Integer;
begin
  if (SkipBlanksLen(ptr,len)=nil)or(ptr^<' ') then Result := ADefaultValue
  else
		begin
			Val(ptr,Result,n);
			if n>0 then
			begin
				Dec(n);
				if n=0 then Result := ADefaultValue;
				Inc(ptr,n);
				Dec(len,n);
			end;
		end;
end;
{---------------------------------------------------------------------------}
function GetNumVariant( var ptr:PChar; var len:Integer ) : Variant;
const Numbers:String = '012345789';
var	i,n : Integer;
		c, first : Char;
begin
	Result := Null;
	if SkipBlanksLen(ptr,len)=nil then Exit;
  first := ptr^;
  if first<>'-' then	i := QScanSkipChars( ptr, len, Numbers, 1 ) {skip}
	else i := QScanSkipChars( ptr+1, len-1, Numbers, 1 ); {skip}
  if i<=0 then
  begin
    Val(ptr,i,n);
    Result := i;
    ptr := nil;
    len := 0;
    Exit;
	end;
	if first<>'-' then Dec(i);
	if i=0 then Exit;
  c:= (ptr+i)^;
  if (first<>'-')and(i<=2)and((c=':')or(c='.')or(c='/')) then
  begin
		n := QScanSkipChars( ptr, len, '012345789:/. ', 1 ); {skip}
    if n=0 then n:=len;
    if (n>=8)and((ptr+i+3)^=c) then
    begin
    	Dec(n);
			Result := StrToDateTime(Copy(ptr,1,n));
	    Inc(ptr,n);
  	  Dec(len,n);
      Exit;
		end;
  end;
  if c='.' then
  begin
    Inc(i);
		n := QScanSkipChars( ptr+i, len-i, Numbers, 1 ); {skip}
    if n=0 then n:=len
		else Dec(n);
    Inc(n,i);
		Result := StrToFloat(Copy(ptr,1,n));
    Inc(ptr,n);
 	  Dec(len,n);
    Exit;
  end;
	Result := StrToInt(Copy(ptr,1,i));
  Inc(ptr,i);
  Dec(len,i);
end;
//=============================================================
//   (  )
function QMemScan( ptr : PChar; len : Integer; c : Char ) : PChar;
asm
	test	eax,eax
  jz		@@Ret
	push	esi
  mov		esi,eax	// esi <- ptr
	mov		eax,ecx
  mov		ah,al		// ah <- al <- c
  mov		ecx,edx	// ecx <- len
	test	ecx,ecx
  jz		@@Ret0
@@Next:	// ???????  ?
	lodsb
	cmp		al,ah
  je		@@Exit
  cmp		al,''''
  je		@@SkipQuoted
  cmp		al,'"'
  je		@@SkipQuoted
  loop	@@Next
	jmp		@@Ret0

@@skipQuoted:
	dec		ecx
	jcxz	@@Ret0
  xchg	esi,edi
	repne	scasb
  xchg	esi,edi
  je		@@Next
@@Ret0:
  xor		esi,esi
@@Exit:
  mov		eax,esi
	pop		esi
@@Ret:
end;

{---------------------------------------------------------------------------}
function QStrScan( str : String; c : Char ) : Integer;
asm
	test		EAX,EAX
  JZ		@@Ret
	PUSH	ESI
  MOV		ESI,EAX
  MOV		ECX,[EAX-4]
	test		ECX,ECX
  JZ		@@Ret0
  XCHG	EDX,EAX
  MOV		AH,AL
@@Next:
	LODSB
	CMP		AL,AH
  JE		@@Exit
  CMP		AL,''''
  JE		@@SkipQuoted
  CMP		AL,'"'
  JE		@@SkipQuoted
  LOOP	@@Next
@@Ret0:
  XOR		EAX,EAX
	JMP		@@RetEAX

@@SkipQuoted:
	DEC		ECX
	JCXZ	@@Ret0
  XCHG	ESI,EDI
	REPNE	SCASB
  XCHG	ESI,EDI
  JNE		@@Ret0
  JMP		@@Next

@@Exit:
  MOV		EAX,ESI
	SUB		EAX,EDX
@@RetEAX:
	POP		ESI
@@Ret:
end;

function QStrSplit( var str : String; c : char ) : String;
var	n, l : Integer;
begin
	n := QStrScan( str, c );
  if n=0 then
   begin
    Result := QStrUnquote( str );
  	str := '';
   end
	else
   begin
		Result := QStrUnquote( Copy( str, 1, n-1 ));
    l := Length(str)-n;
    if l>0 then str := Copy( str, n+1, l )
    else	str := '';
   end;
end;

{---------------------------------------------------------------------------}
function NextToken( var ptr : PChar ) : PChar; assembler;
asm
	OR		EAX,EAX
  JZ		@@Ret
	PUSH	ESI
  MOV		ESI,[EAX]
  OR		ESI,ESI
  JZ    @@RetZero
  PUSH	EAX
@@SkipSpaces1:
	LODSB
	CMP		AL,' '
  JE		@@Next1
  JNC		@@TestComma
  CMP		AL,9
  JNE		@@ControlChar
@@Next1:
  LOOP	@@SkipSpaces1
	JMP		@@Exit0

@@ControlChar:
	OR		EAX,EAX
  JZ		@@RetNil
	CMP		AL,13
  JE		@@RetNil
	CMP		AL,10
  JNE		@@Exit
@@RetNil:
  POP		EAX
  DEC		ESI
  MOV		[EAX],ESI
@@RetZero:
	XOR		EAX,EAX
 	JMP		@@Ret0

@@TestComma:
	CMP		AL,','
  JE    @@SkipSpaces2
	CMP		AL,';'
	JNE		@@Exit
@@SkipSpaces2:
	LODSB
	CMP		AL,' '
  JE		@@Next2
  JNC		@@Exit
  CMP		AL,9
  JNE		@@ControlChar
@@Next2:
  LOOP	@@SkipSpaces2
@@Exit0:
	XOR		ESI,ESI
  JMP		@@RetESI

@@Exit:
	DEC		ESI
@@RetESI:
	MOV		EAX,ESI
  POP		ESI
  MOV		[ESI],EAX
@@Ret0:
	POP		ESI
@@Ret:
end;

function GetTokenInteger(var ptr:PChar):Integer;
var n:Integer;
begin
	ptr := SkipBlanks(ptr);
  if (ptr=nil)or(ptr^<' ') then Result:=0
  else
   begin
    Val(ptr,Result,n);
    ptr:=ptr+n;
		NextToken(ptr);
   end;
end;

function GetIntegers(ptr:PChar;Arg:Array of PInteger):PChar;
var i,n:Integer;
begin
  for i:=Low(Arg) to High(Arg) do
  begin
    Arg[i]^ := GetTokenInteger(ptr);
    if ptr=nil then break;
    Val(ptr,Arg[i]^,n);
    ptr:=ptr+n;
  end;
  Result:=ptr;
end;

//-------------------------------------------------------------
function SpellCurrency( ANum : Currency ) : String;
type
	TUnitGender = ( ugMale, ugFemale );
const
		female : Array[1..2] of String = ('','');
		upto20 : Array[1..20] of String =
			('','','','','','','','','','',
			'','','','','','','','','','');
		decades: Array [2..9] of String = ('','','','','','','','');
		hundreds: Array [1..9] of String = ('','','','','','','','','');
var n, m, i : LongInt;
		s : String;

	function SpellUpTo1000( const AGender : TUnitGender ) : String;
	var	k, h : Integer;
	begin
		k := m mod 100;
		h := m div 100;
		if h=0 then Result := '' else Result := hundreds[h];
		if k>20 then
		begin
			h := k div 10;
			k := k mod 10;
			Result := StrSpaceStr(Result,' ',decades[h]);
		end;
		if k<>0 then
			if (k<3)and(AGender=ugFemale) then Result := StrSpaceStr(Result,' ',female[k])
			else Result := StrSpaceStr(Result,' ',upto20[k]);
	end;

	procedure SpellNum( var Res : String; AUnits : Array of String; const AGender : TUnitGender );
	begin
		i := n mod 100;
		if (i>20)or(i<10) then
			case i mod 10 of
			1: s := AUnits[0];
			2,3,4: s := AUnits[1];
			else
				s := AUnits[2];
			end
		else s := AUnits[2];
		m := n mod 1000;
		n := n div 1000;
		Res := StrSpaceStr( SpellUpTo1000(AGender)+' '+s,' ',Res);
	end;

begin
	n := Trunc(ANum);
	if n=0 then Result := ' '
	else
		begin
			Result := '';
			SpellNum(Result,['','',''],ugMale);
			if n<>0 then
			begin
				SpellNum(Result,['','',''],ugFemale);
				if n<>0 then
				begin
					SpellNum(Result,['','',''],ugMale);
					if n<>0 then SpellNum(Result,['','',''],ugMale);
				end;
			end;
		end;
	n := Trunc(ANum*100) mod 100;
	if n<10 then s := '0'+Chr(Ord('0')+n) else s := IntToStr(n);
	if (n>20)or(n<10) then
		case n mod 10 of
		1: s := s+' ';
		2,3,4: s := s+' ';
		else
			s := s+' ';
		end
	else s := s+' ';
	CharUpperBuff( PChar(Result), 1 );
	Result := Result+' '+s;
end; {SpellCurrency}
//-------------------------------------------------------------
function RoundCents( const ACurr : Currency ) : Currency;
var	D : Currency;
begin
	D := ACurr/100;
	Result := D*100;
end; {RoundCents}
//-------------------------------------------------------------
function SpellDate( const ADate : TDateTime ) : String;
const aMonth : array [1..12] of String = ('','','','','','','','','','','','');
var	wYear, wMonth, wDay : Word;
begin
	DecodeDate( ADate, wYear, wMonth, wDay );
	Result := IntToStr(wDay)+' '+aMonth[wMonth]+' '+IntToStr(wYear);
end; {SpellDate}
//-------------------------------------------------------------
function IncMonth( ADate : TDateTime; const AMonthes : Integer ) : TDateTime;
const aMonth : array [1..12] of Byte = (31,28,31,30,31,30,31,31,30,31,30,31);
var	wYear, wMonth, wDay : Word;
		iMonth : Integer;
begin
	if Trunc(ADate)=0 then ADate := Now;
	DecodeDate( ADate, wYear, wMonth, wDay );
	iMonth := Integer(wMonth-1)+AMonthes;
  if iMonth>=0 then wMonth := (iMonth mod 12) + 1
	else
		begin
			wMonth := (iMonth mod 12) + 13;
      Dec(wYear);
		end;
	if (wDay>aMonth[wMonth])and((wMonth<>2)or(wDay<>29)or((wYear mod 4)<>0)or(wYear=2000)) then	wDay := aMonth[wMonth];
	Inc(wYear,iMonth div 12);
	Result := EncodeDate(wYear,wMonth,wDay)+Frac(ADate);
end; {IncMonth}

end.

